/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// résolution du Sudoku avec le retour sur trace (backtracking)

#include "prelude.h"

int   row(int c) { return c / 9; }
int   col(int c) { return c % 9; }
int group(int c) { return 3 * (row(c) / 3) + col(c) / 3; }

bool same_zone(int c1, int c2) {
  return   row(c1) == row(c2)
      ||   col(c1) == col(c2)
      || group(c1) == group(c2);
}

// vérifie que la valeur à la position p est compatible avec les autres cases
bool check(int grid[81], int p) {
  for (int c = 0; c < 81; c++)
    if (c != p && same_zone(p, c) && grid[p] == grid[c])
      return false;
  return true;
}

// algorithme de retour sur trace (backtracking)
// entrée : grid ne contient pas de contradiction
// sortie : renvoie true si grid a pu être complétée en une solution
//          renvoie false si ce n'est pas possible *et* grid est inchangée
bool solve(int grid[81]) {
  for (int c = 0; c < 81; c++)
    if (grid[c] == 0) {
      for (int v = 1; v <= 9; v++) {
        grid[c] = v;
        if (check(grid, c) && solve(grid))
          return true;
      }
      grid[c] = 0;
      return false;
    }
  return true;
}

// compter les solutions

int count(int grid[81]) {
  for (int c = 0; c < 81; c++)
    if (grid[c] == 0) {
      int s = 0;
      for (int v = 1; v <= 9; v++) {
        grid[c] = v;
        if (check(grid, c))
          s += count(grid);
      }
      grid[c] = 0;
      return s;
    }
  return 1;
}

// pour tester

void print(int grid[81]) {
  for (int i = 0; i < 9; i++) {
    if (i % 3 == 0) printf("+---+---+---+\n");
    for (int j = 0; j < 9; j++) {
      if (j % 3 == 0) printf("|");
      printf("%d", grid[9*i+j]);
    }
    printf("|\n");
  }
  printf("+---+---+---+\n");
}

void solve_string(char *s) {
  int grid[81];
  for (int i = 0; i < 81; i++)
    grid[i] = s[i] - '0';
  solve(grid);
  print(grid);
}
void count_string(char *s) {
  int grid[81];
  for (int i = 0; i < 81; i++)
    grid[i] = s[i] - '0';
  printf("%d solutions\n", count(grid));
}

int main() {
  // l'exemple utilisé dans le livre
  solve_string("200000060000075030048090100000300000300010009000008000001020570080730000090000004");
  /* count_string("200000060000070030048090100000300000300010000000008000001020570080730000090000004"); */
  /* return 0; */
  char line[82];
  FILE *f = fopen("../data/sudoku.txt", "r");
  int count = 0;
  while (fread(line, 1, 82, f) == 82) {
    assert(line[81] == '\n');
    printf("problem %d\n", ++count);
    solve_string(line);
  }
  fclose(f);
  return 0;
}
